
#include <stdint.h>
#include <string.h>
#include "boards.h"
#include "app_util_platform.h"
#include "nrf_gpio.h"
#include "TCA6057.h"
#include "ble_srv_common.h"
#include "app_error.h"
#include "nrf_drv_twi.h"
#include "nrf_delay.h"

/* Indicates if reading operation from accelerometer has ended. */
static volatile bool m_xfer_done = true;
/* Indicates if setting mode operation has ended. */
static volatile bool m_set_mode_done = false;
static volatile bool m_tx_done = false;
static volatile bool m_rx_done = false;
/* TWI instance. */
static const nrf_drv_twi_t m_twi_device = NRF_DRV_TWI_INSTANCE(0);

#define ARDUINO_I2C_SCL_PIN 8
#define ARDUINO_I2C_SDA_PIN 6

static uint8_t m_fromI2Cdevice[20];
static int16_t gyroVals[3];
static int16_t accelVals[3];
//static int16_t magVals[3];
//LED DRIVER ****************************************************************
void     writeTCA(uint8_t* valToWrite, uint8_t numBytesToWrite){
    uint32_t err_code;
    m_tx_done = false;
    err_code = nrf_drv_twi_tx(&m_twi_device, TCA6057_ADDR, (uint8_t*)valToWrite, numBytesToWrite, false);  
    APP_ERROR_CHECK(err_code);
    while(!m_tx_done);
}

uint8_t* readTCA(uint8_t addrToRead,uint8_t numBytesToRead){
    uint32_t err_code;
    writeTCA(&addrToRead, 1);
    m_rx_done = false;
    err_code = nrf_drv_twi_rx(&m_twi_device, TCA6057_ADDR, (uint8_t*)&m_fromI2Cdevice, numBytesToRead);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    return (uint8_t*)&m_fromI2Cdevice;
}

//ACCEL *********************************************************************
void     writeAccel(uint8_t* valToWrite, uint8_t numBytesToWrite){
    uint32_t err_code;
    m_tx_done = false;
    err_code = nrf_drv_twi_tx(&m_twi_device, LSM303_ADDR, (uint8_t*)valToWrite, numBytesToWrite, false);  
    APP_ERROR_CHECK(err_code);
    while(!m_tx_done);
}

uint8_t* readAccel(uint8_t addrToRead,uint8_t numBytesToRead){
    uint32_t err_code;
    writeAccel(&addrToRead, 1);
    //TODO change twi instance name
    m_rx_done = false;
    //uint16_t to = 0;
    err_code = nrf_drv_twi_rx(&m_twi_device, LSM303_ADDR, (uint8_t*)&m_fromI2Cdevice, numBytesToRead);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    // {
    //   nrf_delay_ms(1);
    //   to++;
    //   if(to >= 1000) break;
    // };
    return m_fromI2Cdevice;
}

uint8_t readAccelReg(uint8_t addrToRead){
    uint32_t err_code;
    m_fromI2Cdevice[0] = 0x00;
    writeAccel(&addrToRead, 1);
    //TODO change twi instance name
    m_rx_done = false;
    err_code = nrf_drv_twi_rx(&m_twi_device, LSM303_ADDR, (uint8_t*)&m_fromI2Cdevice, 1);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    //uint16_t to = 0;
    // {
    //   nrf_delay_ms(1);
    //   to++;
    //   if(to >= 1000) break;
    // };
    return m_fromI2Cdevice[0];
}

void enableAccel(){
    
    uint8_t setCtrlReg[2] = {CTRL2,0x00};
    writeAccel(setCtrlReg,2);    
    nrf_delay_ms(10);
    // 0x57 = 0b01010111
    // AODR = 0101 (50 Hz ODR); AZEN = AYEN = AXEN = 1 (all axes enabled)
    setCtrlReg[0] = CTRL1;
    setCtrlReg[1] = 0x57;
    writeAccel(setCtrlReg,2);    
    nrf_delay_ms(10);
    // Magnetometer

    // 0x64 = 0b01100100
    // M_RES = 11 (high resolution mode); M_ODR = 001 (6.25 Hz ODR)
    setCtrlReg[0] = CTRL5;
    setCtrlReg[1] = 0x64;
    writeAccel(setCtrlReg,2);    
    nrf_delay_ms(10);
    
    // 0x20 = 0b00100000
    // MFS = 01 (+/- 4 gauss full scale)
    setCtrlReg[0] = CTRL6;
    setCtrlReg[1] = 0x20;
    writeAccel(setCtrlReg,2);    
    nrf_delay_ms(10);
    
    // 0x00 = 0b00000000
    // MLP = 0 (low power mode off); MD = 00 (continuous-conversion mode)
    setCtrlReg[0] = CTRL7;
    setCtrlReg[1] = 0x00;
    writeAccel(setCtrlReg,2);    
    nrf_delay_ms(10);
}

int16_t* readAccelXYX(){

    uint8_t msbassert = (OUT_X_L_A | (1 << 7));
    uint8_t* results = readAccel(msbassert,6);
    uint8_t xla = results[0];
    uint8_t xha = results[1];
    uint8_t yla = results[2];
    uint8_t yha = results[3];
    uint8_t zla = results[4];
    uint8_t zha = results[5];

    // combine high and low bytes
    // This no longer drops the lowest 4 bits of the readings from the DLH/DLM/DLHC, which are always 0
    // (12-bit resolution, left-aligned). The D has 16-bit resolution
    int16_t ax = (int16_t)(xha << 8 | xla);
    int16_t ay = (int16_t)(yha << 8 | yla);
    int16_t az = (int16_t)(zha << 8 | zla);

    accelVals[0] = ax;
    accelVals[1] = ay;
    accelVals[2] = az;

    return accelVals; 
}

void setClickCnfg (bool enable){
    //TODO Do this better
    if(enable){
        //SET INT1 
        uint8_t setInteruptPin[2] = {CTRL3,0b11000000};
        writeAccel(setInteruptPin,sizeof(setInteruptPin));
        
        //SET HIGH PASS FILTER
         uint8_t setHPF[2] = {CTRL0,0b0000000};
         writeAccel(setHPF,sizeof(setHPF));
        
        //SET XYZ as listening for single click
        uint8_t setXYZListen[2] = {CLICK_CFG,0b00010101};
        writeAccel(setXYZListen,sizeof(setXYZListen));
        
        //SET THRESHOLD
        uint8_t setThresh[2] = {CLICK_THS,0b00000111};
        writeAccel(setThresh,sizeof(setThresh));
        
        //SET INT1 
        uint8_t setTimeLimit[2] = {TIME_LIMIT,0b00111111};
        writeAccel(setTimeLimit,sizeof(setTimeLimit));
        
    }
    else{
        //SET INT1 
        uint8_t setInteruptPinOff[2] = {CTRL3,0b10000000};
        writeAccel(setInteruptPinOff,sizeof(setInteruptPinOff));
        
        //SET XYZ as listening for single click
        uint8_t setXYZListenOff[2] = {CLICK_CFG,0b00000000};
        writeAccel(setXYZListenOff,sizeof(setXYZListenOff));
        
        //SET HIGH PASS FILTER
        uint8_t setHPFOff[2] = {CTRL0,0b00000000};
        writeAccel(setHPFOff,sizeof(setHPFOff));
        
        //SET THRESHOLD
        uint8_t setThreshOff[2] = {CLICK_THS,0b00000000};
        writeAccel(setThreshOff,sizeof(setThreshOff));
        
        uint8_t setEnable[2] = {CLICK_SRC,0b00000000};
        writeAccel(setEnable,sizeof(setEnable));
    }
}
bool checkIntrpt(){
    uint8_t* interruptSource = readAccel(CLICK_SRC,1);
    if((interruptSource[0] >> 7) & 1) return true;
    return false;
}

//GYRO **********************************************************************
void writeGyro(uint8_t* valToWrite, uint8_t numBytesToWrite){
    uint32_t err_code;
    m_tx_done = false;
    err_code = nrf_drv_twi_tx(&m_twi_device, D20_SA0_HIGH_ADDRESS, (uint8_t*)valToWrite, numBytesToWrite, false);  
    APP_ERROR_CHECK(err_code);
    while(!m_tx_done);
}

uint8_t* readGyro(uint8_t addrToRead,uint8_t numBytesToRead){
    uint32_t err_code;
    writeGyro(&addrToRead, 1);
    m_rx_done = false;
    err_code = nrf_drv_twi_rx(&m_twi_device, D20_SA0_HIGH_ADDRESS, (uint8_t*)&m_fromI2Cdevice, numBytesToRead);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    return m_fromI2Cdevice;
}

uint8_t readGyroReg(uint8_t addrToRead){
    uint32_t err_code;
    m_fromI2Cdevice[0] = 0x00;
    writeGyro(&addrToRead, 1);
    m_rx_done = false;
    err_code = nrf_drv_twi_rx(&m_twi_device, D20_SA0_HIGH_ADDRESS, (uint8_t*)&m_fromI2Cdevice, 1);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    return m_fromI2Cdevice[0];
}

void enableGyro(){
    // 0x00 = 0b00000000
    // Low_ODR = 0 (low speed ODR disabled)
    uint8_t setCtrlReg[2] = {LOW_ODR,0x00};
    writeGyro(setCtrlReg, 2);
    nrf_delay_ms(10);
    // 0x6F = 0b01101111
    // DR = 01 (200 Hz ODR); BW = 10 (50 Hz bandwidth); PD = 1 (normal mode); Zen = Yen = Xen = 1 (all axes enabled)
    setCtrlReg[0] = CTRL1G;
    setCtrlReg[1] = 0x6F;
    writeGyro(setCtrlReg,2);//CTRL1G, 0x6F);
    nrf_delay_ms(10);
    // 0x00 = 0b00000000
    // FS = 00 (+/- 250 dps full scale)
    setCtrlReg[0] = CTRL4G;
    setCtrlReg[1] = 0x00;
    writeGyro(setCtrlReg,2);//CTRL4G, 0x00);
    nrf_delay_ms(10);   
}

int16_t* readGyroXYZ(){

    uint8_t msbassert = (OUT_X_L_G | (1 << 7));
    
    uint8_t* results = readGyro(msbassert,6);
    
    uint8_t xla = results[0];
    uint8_t xha = results[1];
    
    uint8_t yla = results[2];
    uint8_t yha = results[3];
    
    uint8_t zla = results[4];
    uint8_t zha = results[5];

    // combine high and low bytes
    // This no longer drops the lowest 4 bits of the readings from the DLH/DLM/DLHC, which are always 0
    // (12-bit resolution, left-aligned). The D has 16-bit resolution
    int16_t gx = (int16_t)(xha << 8 | xla);
    int16_t gy = (int16_t)(yha << 8 | yla);
    int16_t gz = (int16_t)(zha << 8 | zla);
    
    gyroVals[0] = gx;
    gyroVals[1] = gy;
    gyroVals[2] = gz;

    return gyroVals; 
}

//TEMP **********************************************************************

int32_t readTemp(){
    
    uint8_t setCtrlReg[1] = {0xE3};
    uint32_t err_code;
    m_tx_done = false;
    err_code = nrf_drv_twi_tx(&m_twi_device, TEMP_ADDR, (uint8_t*)setCtrlReg, sizeof(setCtrlReg), false);  
    APP_ERROR_CHECK(err_code);
    nrf_delay_ms(10);
    while(!m_tx_done);
    m_rx_done = false;
    //uint16_t to = 0;
    err_code = nrf_drv_twi_rx(&m_twi_device, TEMP_ADDR, (uint8_t*)&m_fromI2Cdevice, 2);
    APP_ERROR_CHECK(err_code);
    while(!m_rx_done);
    uint8_t msb = m_fromI2Cdevice[0];
    uint8_t lsb = m_fromI2Cdevice[1];

    uint16_t val = msb << 8 | lsb;
    float convertVal = (175.26*val) / 65536 - 46.85; 
    return (int32_t)convertVal;
}


//TWI ************************************************************************
/**
 * @brief TWI events handler.
 */
void twi_handler(nrf_drv_twi_evt_t const * p_event, void * p_context){   
    //uint32_t err_code;
    //static uint8_t m_sample;
    
    switch(p_event->type)
    {
        case NRF_DRV_TWI_EVT_DONE:
            if ((p_event->type == NRF_DRV_TWI_EVT_DONE) &&
                (p_event->xfer_desc.type == NRF_DRV_TWI_XFER_TX))
            {
                if(m_tx_done != true)
                {
                    m_tx_done  = true;
                    return;
                }
            }
            else  if ((p_event->type == NRF_DRV_TWI_EVT_DONE) &&
                (p_event->xfer_desc.type == NRF_DRV_TWI_XFER_RX))
            {
                if(m_rx_done != true)
                {
                    m_rx_done  = true;
                    return;
                }

            }
            break;
        default:
            break;        
    }   
}

void twi_init (void){
    ret_code_t err_code;
    
    const nrf_drv_twi_config_t twi_tca6057_config = {
       .scl                = ARDUINO_I2C_SCL_PIN,
       .sda                = ARDUINO_I2C_SDA_PIN,
       .frequency          = NRF_TWI_FREQ_100K,
       .interrupt_priority = APP_IRQ_PRIORITY_HIGH
    };
    
    err_code = nrf_drv_twi_init(&m_twi_device, &twi_tca6057_config, twi_handler, NULL);
    APP_ERROR_CHECK(err_code);
    
    nrf_drv_twi_enable(&m_twi_device);
}

//UTILITIES

uint8_t map16toU8(int16_t sensorVal){
  uint8_t retVal = 0;
  int16_t int16LowerBound = -32000;
  int16_t int16UpperBound = 32001;
  int16_t uint8LowerBound = 0;//delay
  int16_t uint8UpperBound = 255;
  int16_t result =  (sensorVal - int16LowerBound) * (uint8UpperBound - uint8LowerBound) / (int16UpperBound - int16LowerBound) + uint8LowerBound;
  retVal  =  (uint8_t)result;
  return retVal;
}

uint8_t map16toNibble(int16_t sensorVal){
  uint8_t retVal = 0;
  int16_t int16LowerBound = -32000;
  int16_t int16UpperBound = 32001;
  int16_t uint8LowerBound = 0;//delay
  int16_t uint8UpperBound = 15;
  int16_t result =  (sensorVal - int16LowerBound) * (uint8UpperBound - uint8LowerBound) / (int16UpperBound - int16LowerBound) + uint8LowerBound;
  retVal  =  (uint8_t)result;
  return retVal;
}